{$N+,G+}
UNIT Graphics;

INTERFACE

USES Dos, Crt;
PROCEDURE InitVGA;
PROCEDURE InitSVGA(Mode : Word);
PROCEDURE CloseGraphics;
PROCEDURE WaitRetrace;
PROCEDURE CLRVID(Where : Word;Color : Byte);
FUNCTION  GetVideoPages(NumPages : Byte) : Byte;
PROCEDURE DisposeVideoPages;
PROCEDURE CopyPage(Source,Dest : Word);
PROCEDURE PutPixel(X,Y : LongInt;Color, Method : Byte;Where : Word);
PROCEDURE PutPixelW(X,Y : LongInt;Color : Byte;Where : Word);
PROCEDURE Line(X1,Y1,X2,Y2 : Integer;Color : Byte;Method : Byte;Where : Word);
PROCEDURE VLine(Y1,Y2,X : Integer;Color : Byte;Method : Byte;Where : Word);
PROCEDURE HLine(X1,X2,Y : Integer;Color : Byte;Method : Byte;Where : Word);
PROCEDURE Rectangle(X1, Y1, X2, Y2 : Integer; Color : Byte;Method : Byte;Where : Word);
PROCEDURE RectangleR(X1, Y1, X2, Y2 : Integer; Rot : Byte; Color : Byte;Method : Byte;Where : Word);
PROCEDURE RRectangle(X1, Y1, X2, Y2 : Integer; Color : Byte;Method : Byte;Where : Word);
PROCEDURE FillRectangle(X1, Y1, X2, Y2 : Integer;Color, Fill : Byte;Method : Byte;Where : Word);
PROCEDURE FillRRectangle(X1, Y1, X2, Y2 : Integer;Color, Fill : Byte;Method : Byte;Where : Word);
PROCEDURE Circle(X, Y, Rad : Word; Color, Method : Byte; Where : Word);
PROCEDURE FillCircle(X, Y, Rad : Word; Color, Fill, Method : Byte;Where : Word);
PROCEDURE Ellipse(X, Y, XRad, YRad : Word; Rot : Byte; Color, Method : Byte; Where : Word);
PROCEDURE SetColor(Color, R, G, B : Byte);
PROCEDURE GetColor(Color : Byte; VAR R, G, B : Byte);
PROCEDURE WriteChar(xc, yc : Integer; c : Char; color, transparent, width : Integer; Where : Word);
PROCEDURE WriteText(x, y, color : Integer; s : string; transparent, width : Integer; Where : Word);
FUNCTION  Sqrt(Num : Double) : Real;

CONST
     VGA = $A000;
     ROM_CHAR_SET_SEG : Word = $F000;
     ROM_CHAR_SET_OFF : Word = $FA6E;
     rom_char_set : Pointer = Ptr($F000,$FA6E);
     ROM_CHAR_HEIGHT : Byte = 8;
     ROM_CHAR_WIDTH  : Byte = 8;

TYPE
    VPage = Array[0..63999] Of Byte;
    VPagePtr = ^VPage;

VAR
   VAddr1,
   VAddr2,
   VAddr3,
   VAddr4        : Word;
   VirScr1,
   VirScr2,
   VirScr3,
   VirScr4       : VPagePtr;
   ScrHeight,
   ScrWidth         : Word;
{   SinLook       : Array[0..255] Of Real;
   CosLook       : Array[0..255] Of Real;}
   SinLook       : Array[0..255] Of Integer;
   CosLook       : Array[0..255] Of Integer;

IMPLEMENTATION

VAR
   CharSet       : Array[0..255,0..15] Of Byte;
   OldBank       : Byte;
   NumVidPages   : Byte;
   F             : File;
   FontSeg,
   FontOfs       : Word;

PROCEDURE InitVGA; Assembler;
ASM
   Mov Ax, 13h
   Int 10h
   Mov ScrWidth, 320
   Mov ScrHeight, 200
END;

PROCEDURE InitSVGA; Assembler;
ASM
   Mov Ax, $4F02
   Mov Bx, Mode
   Int 10h
   Cmp Mode, 100h
   Jb @Error
   Ja @Check640x480
   Mov ScrWidth, 640
   Mov ScrHeight, 400
   Jmp @Error
  @Check640x480:
   Cmp Mode, 102h
   Ja @Check800x600
   Mov ScrWidth, 640
   Mov ScrHeight, 480
   Jmp @Error
  @Check800x600:
   Cmp Mode, 104h
   Ja @Check1024x768
   Mov ScrWidth, 800
   Mov ScrHeight, 600
   Jmp @Error
  @Check1024x768:
   Mov ScrWidth, 1024
   Mov ScrHeight, 768
   Jmp @Error
  @Error:
END;

PROCEDURE CloseGraphics; Assembler;
ASM
   Mov Ax, 03h
   Int 10h
END;

PROCEDURE WaitRetrace; Assembler;
ASM
    Mov dx,3DAh
@l1:
    In al,dx
    And al,08h
    Jnz @l1
@l2:
    In al,dx
    And al,08h
    Jz  @l2
END;

PROCEDURE CLRVID; Assembler;
ASM
   Mov Di, Where
   Mov Es, Di
   Xor Di, Di
   Mov Al, Color
   Mov Ah, Al
   Mov Cx, 32000
   REP STOSW
END;

FUNCTION GetVideoPages;
Begin
     If MaxAvail < 64000 Then Begin GetVideoPages := 0; Exit; End;
     If (MaxAvail > 256000) and (NumPages >= 4) Then
        Begin
             New(VirScr4);VAddr4 := Seg(VirScr4^);
             New(VirScr3);VAddr3 := Seg(VirScr3^);
             New(VirScr2);VAddr2 := Seg(VirScr2^);
             New(VirScr1);VAddr1 := Seg(VirScr1^);
             NumVidPages := 4;
        End ELSE
     If (MaxAvail > 192000) and (NumPages >= 3) Then
        Begin
             New(VirScr3);VAddr3 := Seg(VirScr3^);
             New(VirScr2);VAddr2 := Seg(VirScr2^);
             New(VirScr1);VAddr1 := Seg(VirScr1^);
             NumVidPages := 3;
        End ELSE
     If (MaxAvail > 128000) and (NumPages >= 2)  Then
        Begin
             New(VirScr2);VAddr2 := Seg(VirScr2^);
             New(VirScr1);VAddr1 := Seg(VirScr1^);
             NumVidPages := 2;
        End ELSE
     If (MaxAvail >  64000) and (NumPages >= 1)  Then
        Begin
             New(VirScr1);VAddr1 := Seg(VirScr1^);
             NumVidPages := 1;
        End;
     GetVideoPages := NumVidPages;
End;

PROCEDURE DisposeVideoPages;
Begin
     If NumVidPages > 3 Then Dispose(VirScr4);
     If NumVidPages > 2 Then Dispose(VirScr3);
     If NumVidPages > 1 Then Dispose(VirScr2);
     If NumVidPages > 0 Then Dispose(VirScr1);
End;

PROCEDURE CopyPage; Assembler;
ASM
   Push Ds
   Cld
   Mov Si, Source
   Mov Ds, Si
   Xor Si, Si
   Mov Di, Dest
   Mov Es, Di
   Xor Di, Di
   Mov Cx, 32000
  @Transfer:
   LODSW
   STOSW
   Loop @Transfer
   Pop Ds
END;

PROCEDURE PutPixel(X,Y : LongInt;Color, Method : Byte;Where : Word);
VAR Bank : Word;
Begin
 If ScrWidth <= 320 Then
  Begin
     If Method = 0 Then
        Mem[Where:X+Y*ScrWidth] := Color
     ELSE If Method = 1 Then
        Mem[Where:X+Y*ScrWidth] := Mem[Where:X+Y*ScrWidth] XOR Color
     ELSE If Method = 2 Then
        Mem[Where:X+Y*ScrWidth] := Mem[Where:X+Y*ScrWidth] + Color
     ELSE
        Mem[Where:X+Y*ScrWidth] := Mem[Where:X+Y*ScrWidth] - Color;
  End
 ELSE
  Begin
        Bank := (Y*ScrWidth+X) Shr 16;
        If OldBank <> Bank Then
        Begin
             ASM
                Mov Ax, $4f05
                Xor Bh, Bh
                Mov Dx, Bank
                Int $10
             END;
             OldBank := Bank;
        End;
        If Method = 0 Then
           Mem[Where:X+Y*ScrWidth-(Bank Shl 16)] := Color
        ELSE If Method = 1 Then
           Mem[Where:X+Y*ScrWidth-(Bank Shl 16)] := Mem[Where:X+Y*ScrWidth-(Bank Shl 16)] XOR Color
        ELSE If Method = 2 Then
           Mem[Where:X+Y*ScrWidth-(Bank Shl 16)] := Mem[Where:X+Y*ScrWidth-(Bank Shl 16)] + Color
        ELSE
           Mem[Where:X+Y*ScrWidth-(Bank Shl 16)] := Mem[Where:X+Y*ScrWidth-(Bank Shl 16)] - Color;
  End;
End;

PROCEDURE PutPixelW(X,Y : LongInt;Color : Byte;Where : Word);
Begin
     MemW[Where:X+Y*ScrWidth] := Color+Color SHL 8;
End;

PROCEDURE Line(X1,Y1,X2,Y2 : Integer;Color : Byte;Method : Byte;Where : Word);
  { This draws a solid line from X1,Y1 to X2,Y2 in COLOR }
  FUNCTION SGN(X1 : Real) : Integer;
  Begin
       if X1>0 then SGN :=  1;
       if X1<0 then SGN := -1;
       if X1=0 then SGN :=  0;
  End;
VAR
   i, s,
   d1x, d1y, d2x, d2y,
   u, v, m, n           : Integer;
Begin
     u:= X2 - X1;
     v:= Y2 - Y1;
     d1x:= SGN(u);
     d1y:= SGN(v);
     d2x:= SGN(u);
     d2y:= 0;
     m:= ABS(u);
     n := ABS(v);
     If NOT(M > N) Then
     Begin
          d2x := 0 ;
          d2y := SGN(v);
          m := ABS(v);
          n := ABS(u);
     End;
     s := m SHR 1;
     For i := 0 to m Do
     Begin
          If (X1 >= 0) and (X1 < ScrWidth) and (Y1 >= 0) and (Y1 < ScrHeight) then
             PutPixel(X1,Y1,Color,Method,Where);
          s := s + n;
          If NOT(s < m) Then
          Begin
               s := s - m;
               X1:= X1 + d1x;
               Y1 := Y1 + d1y;
          End
          ELSE
          Begin
               X1 := X1 + d2x;
               Y1 := Y1 + d2y;
          End;
     End;
End;

PROCEDURE VLine(Y1,Y2,X : Integer;Color : Byte;Method : Byte;Where : Word); Assembler;
     ASM
        Mov Bx, Where           { ES = Where;           }
        Mov Es, Bx
        Mov Al, Color           { AL = Color;           }
        Mov Cx, X               { CX = X;               }
        Mov Dx, Y1              { DX = Y;               }
        @YCOORDINATE:
        Push Ax                 { Save Color            }
        Push Dx                 { Save Y                }
        Mov Ax, Dx              { AX = Y                }
        Mul ScrWidth            { AX = Y*ScrWidth       }
        Mov Bx, Ax              { BX = Y*ScrWidth       }
        Pop Dx                  { DX = Y                }
        Pop Ax                  { AX = Color            }
        Add Bx, Cx              { BX = Y*ScrWidth+X     }
        Cmp Method, 1           { Check Method          }
        Jb  @Normal             { METHOD = 0-> Normal   }
        Ja  @Subtract           { METHOD = 2-> Subtract }
        Add [ES:Bx], Al         { METHOD = 1-> Add      }
        Jmp @Continue
       @Subtract:
        Sub ES:[Bx], Al
        Jmp @Continue
       @Normal:
        Mov [ES:Bx], Al
       @Continue:
        INC Dx                  { Y = Y + 1             }
        CMP Dx, Y2              { Y = Y2?               }
        JNA @YCOORDINATE
     END;

PROCEDURE HLine(X1,X2,Y : Integer;Color : Byte;Method : Byte;Where : Word); Assembler;
     ASM
        Mov Bx, Where
        Mov Es, Bx
        Mov Cx, X1
        Mov Dx, Y
        Mov Ax, Dx
        Mul ScrWidth
        Mov Bx, Ax
        Mov Al, Color
        @XCOORDINATE:
        Push Bx
        Add Bx, Cx
        Cmp Method, 1
        Jb  @Normal
        Ja  @Subtract
        Add ES:[Bx], Al
        Jmp @Continue
       @Subtract:
        Sub ES:[Bx], Al
        Jmp @Continue
       @Normal:
        Mov ES:[Bx], Al
       @Continue:
        Inc Cx
        Pop Bx
        Cmp Cx, X2
        Jna @XCOORDINATE
     END;

PROCEDURE Rectangle(X1, Y1, X2, Y2 : Integer; Color : Byte;Method : Byte;Where : Word);
     Begin
          Line( X1, Y1, X2, Y1, Color,Method,Where);
          Line( X1, Y2, X2, Y2, Color,Method,Where);
          Line( X1, Y1, X1, Y2, Color,Method,Where);
          Line( X2, Y1, X2, Y2, Color,Method,Where);
     End;

PROCEDURE RectangleR(X1, Y1, X2, Y2 : Integer; Rot : Byte; Color : Byte;Method : Byte;Where : Word);
VAR XT, YT : Integer;
    XS, YS : Integer;
     Begin
          XT := (X2-X1) SHR 1;    YT := (Y2-Y1) SHR 1;
          XS := X1+XT;            YS := Y1+YT;
          Line( XS+ ((-XT)*CosLook[Rot] - (-YT)*SinLook[Rot]) DIV 256,
                YS+ ((-XT)*SinLook[Rot] + (-YT)*CosLook[Rot]) DIV 256,
                XS+ ((+XT)*CosLook[Rot] - (-YT)*SinLook[Rot]) DIV 256,
                YS+ ((+XT)*SinLook[Rot] + (-YT)*CosLook[Rot]) DIV 256, Color,Method,Where);
          Line( XS+ ((-XT)*CosLook[Rot] - (+YT)*SinLook[Rot]) DIV 256,
                1+YS+ ((-XT)*SinLook[Rot] + (+YT)*CosLook[Rot]) DIV 256,
                1+XS+ ((+XT)*CosLook[Rot] - (+YT)*SinLook[Rot]) DIV 256,
                1+YS+ ((+XT)*SinLook[Rot] + (+YT)*CosLook[Rot]) DIV 256, Color,Method,Where);
          Line( XS+ ((-XT)*CosLook[Rot] - (-YT)*SinLook[Rot]) DIV 256,
                YS+ ((-XT)*SinLook[Rot] + (-YT)*CosLook[Rot]) DIV 256,
                XS+ ((-XT)*CosLook[Rot] - (+YT)*SinLook[Rot]) DIV 256,
                YS+ ((-XT)*SinLook[Rot] + (+YT)*CosLook[Rot]) DIV 256, Color,Method,Where);
          Line( 1+XS+ ((+XT)*CosLook[Rot] - (-YT)*SinLook[Rot]) DIV 256,
                YS+ ((+XT)*SinLook[Rot] + (-YT)*CosLook[Rot]) DIV 256,
                1+XS+ ((+XT)*CosLook[Rot] - (+YT)*SinLook[Rot]) DIV 256,
                1+YS+ ((+XT)*SinLook[Rot] + (+YT)*CosLook[Rot]) DIV 256, Color,Method,Where);
     End;

PROCEDURE RRectangle(X1, Y1, X2, Y2 : Integer; Color : Byte;Method : Byte;Where : Word);
     Begin
          VLine(Y1+1,Y2-1,X1,Color,Method,Where);
          VLine(Y1+1,Y2-1,X2,Color,Method,Where);
          HLine(X1+1,X2-1,Y1,Color,Method,Where);
          HLine(X1+1,X2-1,Y2,Color,Method,Where);
     End;

PROCEDURE FillRectangle(X1, Y1, X2, Y2 : Integer;Color, Fill : Byte;Method : Byte;Where : Word);
     VAR
        X, Y : Integer;
     Begin
          If (Method = 1) or (Method = 3) Then
             Rectangle(X1,Y1,X2,Y2,Color,1,Where)
          ELSE If (Method = 4) or (Method = 6) Then
             Rectangle(X1,Y1,X2,Y2,Color,2,Where)
          ELSE If (Method = 2) or (Method = 5) Then
             Rectangle(X1,Y1,X2,Y2,Color,0,Where)
          ELSE
             Rectangle(X1,Y1,X2,Y2,Color,0,Where);
          If (Method = 5) or (Method = 6) Then Method := 2;
          If (Method = 1) or (Method = 4) Then Method := 1;
          For X := X1+1 to X2-1 Do
              VLine(Y1+1,Y2-1,X,Fill,Method,Where);
     End;

PROCEDURE FillRRectangle(X1, Y1, X2, Y2 : Integer;Color, Fill : Byte;Method : Byte;Where : Word);
     VAR
        X, Y : Integer;
     Begin
          If (Method = 1) or (Method = 3) Then
             RRectangle(X1,Y1,X2,Y2,Color,1,Where)
          ELSE If (Method = 4) or (Method = 6) Then
             RRectangle(X1,Y1,X2,Y2,Color,2,Where)
          ELSE If (Method = 2) or (Method = 5) Then
             RRectangle(X1,Y1,X2,Y2,Color,0,Where)
          ELSE
             RRectangle(X1,Y1,X2,Y2,Color,0,Where);
          If (Method = 5) or (Method = 6) Then Method := 2;
          If (Method = 1) or (Method = 4) Then Method := 0;
          For X := X1+1 to X2-1 Do
              VLine(Y1+1,Y2-1,X,Fill,Method,Where);
     End;

PROCEDURE Circle(X, Y, Rad : Word; Color, Method : Byte; Where : Word);
VAR
   W  : Byte;
   Z  : Real;
   Z2 : Real;

Begin
     For W := 0 to 255 Do
         Begin
              Z  := (Rad*SinLook[W])/2.85;
              Z2 := (Rad*CosLook[W])/2.85;
              PutPixel(X+Round(Z-Z2),Y+Round(Z+Z2),Color,Method,Where);
         End;
End;

PROCEDURE FillCircle(X, Y, Rad : Word; Color, Fill, Method : Byte;Where : Word);
VAR
   W  : Byte;
   W2 : Integer;
   Z  : Real;
   Z2 : Real;

Begin
     For W := 0 to 255 Do
         Begin
              If Rad > 2 Then
                 Begin
                      Z  := ((Rad-2)*SinLook[W])/2.85;
                      Z2 := ((Rad-2)*CosLook[W])/2.85;
                      For W2 := 0-Round(Z+Z2) to Round(Z+Z2) Do
                          PutPixel(X+W2,Y+Round(Z-Z2),Fill,Method,Where);
                 End;
              Z  := (Rad*SinLook[W])/2.85;
              Z2 := (Rad*CosLook[W])/2.85;
              PutPixel(X+Round(Z+Z2),Y+Round(Z-Z2),Color,Method,Where);
         End;
End;

PROCEDURE Ellipse(X, Y, XRad, YRad : Word; Rot : Byte; Color, Method : Byte; Where : Word);
VAR
   W  : Byte;
   Z  : Integer;
   Z2 : Integer;
   OldX : Word;
   OldY : Word;

Begin
     Z  := (XRad*SinLook[255] SHR 8)*CosLook[Rot] SHR 8 - (YRad*CosLook[255] SHR 8)*SinLook[Rot] SHR 8;
     Z2 := (XRad*SinLook[255] SHR 8)*SinLook[Rot] SHR 8 + (YRad*CosLook[255] SHR 8)*CosLook[Rot] SHR 8;
     For W := 0 to 255 Do
         Begin
              OldX := Z; OldY := Z2;
              Z  := ((XRad*SinLook[W] SHR 8)*CosLook[Rot] - (YRad*CosLook[W] SHR 8)*SinLook[Rot]) SHR 8;
              Z2 := ((XRad*SinLook[W] SHR 8)*SinLook[Rot] + (YRad*CosLook[W] SHR 8)*CosLook[Rot]) SHR 8;
              Line(X+Z,Y+Z2,X+OldX,Y+OldY,Color,Method,Where);
         End;
End;

PROCEDURE SetColor(Color, R, G, B : Byte); Assembler;
ASM
   Mov Dx, 3c8h
   Mov Al, Color
   Out Dx, Al
   Inc Dx
   Mov Al, R
   Out Dx, Al
   Mov Al, G
   Out Dx, Al
   Mov Al, B
   Out Dx, Al
END;

PROCEDURE GetColor(Color : Byte; VAR R, G, B : Byte);
Begin
     Port[$3c7] := Color;
     R := Port[$3c9];
     G := Port[$3c9];
     B := Port[$3c9];
End;

{PROCEDURE GetColor(Color : Byte; VAR R, G, B : Byte); Assembler;
ASM
   Mov Dx, 3c7h
   Mov Al, Color
   Out Dx, Al
   Mov Dx, 3c9h
   In
END;}

PROCEDURE WriteChar(xc, yc : Integer; c : Char; color, transparent, width : Integer; Where : Word);
VAR
   offset       : Word;
   x, y         : Integer;
   work_char    : ^char;
   bit_mask     : byte;
Begin
     work_char := ptr(ROM_CHAR_SET_SEG,ROM_CHAR_SET_OFF + ord(c)*ROM_CHAR_HEIGHT);

     offset := yc * ScrWidth + xc;

     For y := 0 to ROM_CHAR_HEIGHT-1 do
     Begin
          bit_mask := $80;
          If Width = 0 Then
          For x := 0 to ROM_CHAR_WIDTH-1 do
          Begin
               if (ord(work_char^) and bit_mask) <> 0 then
                  mem[Where:offset+x] := color
               else
               if (transparent = 0) then
                  mem[Where:offset+x] := 0;

               bit_mask := bit_mask shr 1;
          End
          ELSE
          For x := 0 to ROM_CHAR_WIDTH-1 do
          Begin
               if ((ord(work_char^) and bit_mask) <> 0) and (x mod (width+1) = 0) then
                  mem[Where:offset+x shr width] := color OR Color*byte(ord(work_char^) and (bit_mask shr 1) <> 0)
               else
               if (transparent = 0) then
                  mem[Where:offset+x shr width] := 0;

               bit_mask := bit_mask shr 1;
          End;

          offset := offset + ScrWidth;
          work_char := ptr(seg(work_char^),ofs(work_char^)+1);
     End;
End;

PROCEDURE WriteText(x, y, color : Integer; s : string; transparent, width : Integer; Where : Word);
VAR
   index : Integer;
Begin
     for index := 0 to Length(s)-1 do
     Begin
         WriteChar(x,y,s[index+1],color,transparent, width, Where);
         x := x + ROM_CHAR_WIDTH SHR Width + Width;
     End;
End;

FUNCTION Sqrt(Num : Double) : Real;   {23 Clock Ticks, CRT.TPU uses 27}
CONST CG : Double = 1.0;
VAR NewGuess : Double;
Begin
     While (ABS(CG-NewGuess) > 0.00005) Do
     Begin
          CG := NewGuess;
          NewGuess := 0.5*(CG+Num/CG)
     End;
     Sqrt := NewGuess;
End;

Begin
{     Assign(F,'GAMEFONT.FNT');Reset(F,1);
     BlockRead(F,CharSet,SizeOf(CharSet));
     Close(F);}
{     asm
        Mov Ax, $1130
        Mov Bh, $06
        Int $10
        Mov FontSeg, Es
        Mov FontOfs, Bp
     end;}
     For OldBank := 0 to 255 Do
         Begin
              SinLook[OldBank] := Round(Sin(2*PI*OldBank/256)*256);
              CosLook[OldBank] := Round(Cos(2*PI*OldBank/256)*256);
         End;
End.